import os
import pytest
from .util import (
    assert_has_threads,
    get_querier,
    get_write_manifest,
    ProjectDefinition,
)


@pytest.mark.supported('postgres')
def test_rpc_compile_threads(
    project_root, profiles_root, dbt_profile, unique_schema
):
    project = ProjectDefinition(
        models={'my_model.sql': 'select 1 as id'}
    )
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )
    with querier_ctx as querier:
        results = querier.async_wait_for_result(querier.compile(threads=5))
        assert_has_threads(results, 5)

        results = querier.async_wait_for_result(
            querier.cli_args('compile --threads=7')
        )
        assert_has_threads(results, 7)


@pytest.mark.supported('postgres')
def test_rpc_compile_state(
    project_root, profiles_root, dbt_profile, unique_schema
):
    project = ProjectDefinition(models={'my_model.sql': 'select 1 as id'})
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )
    with querier_ctx as querier:
        state_dir = os.path.join(project_root, 'state')
        os.makedirs(state_dir)

        results = querier.async_wait_for_result(
            querier.compile()
        )
        assert len(results['results']) == 1

        get_write_manifest(querier, os.path.join(state_dir, 'manifest.json'))

        project.models['my_model.sql'] = 'select 2 as id'
        project.write_models(project_root, remove=True)

        querier.sighup()
        assert querier.wait_for_status('ready') is True

        results = querier.async_wait_for_result(
            querier.compile(state='./state', models=['state:modified'])
        )
        assert len(results['results']) == 1

        get_write_manifest(querier, os.path.join(state_dir, 'manifest.json'))

        results = querier.async_wait_for_result(
            querier.compile(state='./state', models=['state:modified']),
        )
        assert len(results['results']) == 0


@pytest.mark.supported('postgres')
def test_rpc_compile_sql_disabled_node(
    project_root, profiles_root, dbt_profile, unique_schema
):
    project = ProjectDefinition()
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )
    with querier_ctx as querier:

        result = querier.async_wait_for_error(
            querier.compile_sql('{{ config(enabled=false) }}\n select 1 as id')
        )

        assert 'Trying to compile a node that is disabled' \
            in result['data']['message']



import base64
import signal
import threading
from datetime import datetime
from typing import Dict, Any

from dbt.flags import env_set_truthy
from dbt.adapters.factory import get_adapter
from dbt.clients.jinja import extract_toplevel_blocks
from dbt.config.runtime import RuntimeConfig
from dbt.contracts.graph.manifest import Manifest
from dbt.contracts.graph.nodes import RPCNode
from dbt_rpc.contracts.rpc import RPCExecParameters
from dbt_rpc.contracts.rpc import RemoteExecutionResult
from dbt.exceptions import RPCKilledException, DbtInternalError
from dbt.logger import GLOBAL_LOGGER as logger
from dbt.parser.manifest import process_node, process_macro
from dbt_rpc.parser.rpc import RPCCallParser, RPCMacroParser
from dbt_rpc.rpc.error import invalid_params
from dbt_rpc.rpc.node_runners import RPCCompileRunner, RPCExecuteRunner
from dbt.task.compile import CompileTask
from dbt.task.run import RunTask

from .base import RPCTask


SINGLE_THREADED_HANDLER = env_set_truthy('DBT_SINGLE_THREADED_HANDLER')


def add_new_refs(
    manifest: Manifest,
    config: RuntimeConfig,
    node: RPCNode,
    macros: Dict[str, Any]
) -> None:
    """Given a new node that is not in the manifest, insert the new node
    into it as if it were part of regular ref processing.
    """
    if config.args.single_threaded or SINGLE_THREADED_HANDLER:
        manifest = manifest.deepcopy()
    # it's ok for macros to silently override a local project macro name
    manifest.macros.update(macros)

    for macro in macros.values():
        process_macro(config, manifest, macro)

    # We used to do 'manifest.add_nodes({node.unique_id: node}) here, but the
    # node has already been added to the Manifest by the RPCCallParser
    # now that we save nodes to the Manifest instead of ParseResults.
    process_node(config, manifest, node)


class RemoteRunSQLTask(RPCTask[RPCExecParameters]):
    def runtime_cleanup(self, selected_uids):
        """Do some pre-run cleanup that is usually performed in Task __init__.
        """
        self.run_count = 0
        self.num_nodes = len(selected_uids)
        self.node_results = []
        self._skipped_children = {}
        self._skipped_children = {}
        self._raise_next_tick = None

    def decode_sql(self, sql: str) -> str:
        """Base64 decode a string. This should only be used for sql in calls.

        :param str sql: The base64 encoded form of the original utf-8 string
        :return str: The decoded utf-8 string
        """
        # JSON is defined as using "unicode", we'll go a step further and
        # mandate utf-8 (though for the base64 part, it doesn't really matter!)
        base64_sql_bytes = str(sql).encode('utf-8')

        try:
            sql_bytes = base64.b64decode(base64_sql_bytes, validate=True)
        except ValueError:
            self.raise_invalid_base64(sql)

        return sql_bytes.decode('utf-8')

    @staticmethod
    def raise_invalid_base64(sql):
        raise invalid_params(
            data={
                'message': 'invalid base64-encoded sql input',
                'sql': str(sql),
            }
        )

    def _extract_request_data(self, data):
        data = self.decode_sql(data)
        macro_blocks = []
        data_chunks = []
        for block in extract_toplevel_blocks(data):
            if block.block_type_name == 'macro':
                macro_blocks.append(block.full_block)
            else:
                data_chunks.append(block.full_block)
        macros = '\n'.join(macro_blocks)
        sql = ''.join(data_chunks)
        return sql, macros

    def _get_exec_node(self):
        if self.manifest is None:
            raise DbtInternalError(
                'manifest not set in _get_exec_node'
            )

        macro_overrides = {}
        macros = self.args.macros
        sql, macros = self._extract_request_data(self.args.sql)

        if macros:
            macro_parser = RPCMacroParser(self.config, self.manifest)
            for node in macro_parser.parse_remote(macros):
                macro_overrides[node.unique_id] = node

        self.manifest.macros.update(macro_overrides)
        rpc_parser = RPCCallParser(
            project=self.config,
            manifest=self.manifest,
            root_project=self.config,
        )

        rpc_node = rpc_parser.parse_remote(sql, self.args.name, self.args.language)
        add_new_refs(
            manifest=self.manifest,
            config=self.config,
            node=rpc_node,
            macros=macro_overrides
        )

        # don't write our new, weird manifest!
        adapter = get_adapter(self.config)
        compiler = adapter.get_compiler()
        self.graph = compiler.compile(self.manifest, write=False)
        # previously, this compiled the ancestors, but they are compiled at
        # runtime now.
        return rpc_node

    def _raise_set_error(self):
        if self._raise_next_tick is not None:
            raise self._raise_next_tick

    def _in_thread(self, node, thread_done):
        runner = self.get_runner(node)
        try:
            self.node_results.append(runner.safe_run(self.manifest))
        except Exception as exc:
            logger.debug('Got exception {}'.format(exc), exc_info=True)
            self._raise_next_tick = exc
        finally:
            thread_done.set()

    def set_args(self, params: RPCExecParameters):
        self.args.name = params.name
        self.args.sql = params.sql
        self.args.macros = params.macros
        self.args.language = params.language

    def handle_request(self) -> RemoteExecutionResult:
        # we could get a ctrl+c at any time, including during parsing.
        thread = None
        started = datetime.utcnow()
        try:
            node = self._get_exec_node()

            selected_uids = [node.unique_id]
            self.runtime_cleanup(selected_uids)

            thread_done = threading.Event()
            thread = threading.Thread(target=self._in_thread,
                                      args=(node, thread_done))
            thread.start()
            thread_done.wait()
        except KeyboardInterrupt:
            adapter = get_adapter(self.config)  # type: ignore
            if adapter.is_cancelable():

                for conn_name in adapter.cancel_open_connections():
                    logger.debug('canceled query {}'.format(conn_name))
                if thread:
                    thread.join()
            else:
                msg = ("The {} adapter does not support query "
                       "cancellation. Some queries may still be "
                       "running!".format(adapter.type()))

                logger.debug(msg)

            raise RPCKilledException(signal.SIGINT)

        self._raise_set_error()

        ended = datetime.utcnow()
        elapsed = (ended - started).total_seconds()
        return self.get_result(
            results=self.node_results,
            elapsed_time=elapsed,
            generated_at=ended,
        )

    def interpret_results(self, results):
        return True


class RemoteCompileTask(RemoteRunSQLTask, CompileTask):
    METHOD_NAME = 'compile_sql'

    def get_runner_type(self, _):
        return RPCCompileRunner


class RemoteRunTask(RemoteRunSQLTask, RunTask):
    METHOD_NAME = 'run_sql'

    def get_runner_type(self, _):
        return RPCExecuteRunner



import pytest

from .util import (
    get_querier,
    ProjectDefinition,
)


@pytest.mark.supported("any")
def test_rpc_reload(project_root, profiles_root, dbt_profile, unique_schema):
    project = ProjectDefinition(models={"my_model.sql": "select 1 as id"})
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )
    with querier_ctx as querier:
        querier.request(method="reload")



import pytest
import os
from .util import (
    assert_has_threads,
    get_querier,
    get_write_manifest,
    ProjectDefinition,
)


@pytest.mark.supported('postgres')
def test_rpc_seed_threads(
    project_root, profiles_root, dbt_profile, unique_schema
):
    project = ProjectDefinition(
        project_data={'seeds': {'+quote_columns': False}},
        seeds={'data.csv': 'a,b\n1,hello\n2,goodbye'},
    )
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )

    with querier_ctx as querier:
        results = querier.async_wait_for_result(querier.seed(threads=5))
        assert_has_threads(results, 5)

        results = querier.async_wait_for_result(
            querier.cli_args('seed --threads=7')
        )
        assert_has_threads(results, 7)


@pytest.mark.supported('postgres')
def test_rpc_seed_include_exclude(
    project_root, profiles_root, dbt_profile, unique_schema
):
    project = ProjectDefinition(
        project_data={'seeds': {'+quote_columns': False}},
        seeds={
            'data_1.csv': 'a,b\n1,hello\n2,goodbye',
            'data_2.csv': 'a,b\n1,data',
        },
    )
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )

    with querier_ctx as querier:
        results = querier.async_wait_for_result(querier.seed(select=['data_1']))
        assert len(results['results']) == 1
        results = querier.async_wait_for_result(querier.seed(select='data_1'))
        assert len(results['results']) == 1
        results = querier.async_wait_for_result(querier.cli_args('seed --select=data_1'))
        assert len(results['results']) == 1

        results = querier.async_wait_for_result(querier.seed(exclude=['data_2']))
        assert len(results['results']) == 1
        results = querier.async_wait_for_result(querier.seed(exclude='data_2'))
        assert len(results['results']) == 1
        results = querier.async_wait_for_result(querier.cli_args('seed --exclude=data_2'))
        assert len(results['results']) == 1


@pytest.mark.supported('postgres')
def test_rpc_seed_state(
    project_root, profiles_root, dbt_profile, unique_schema
):
    project = ProjectDefinition(seeds={'my_seed.csv': 'a,b\n1,hello'})
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )
    with querier_ctx as querier:
        state_dir = os.path.join(project_root, 'state')
        os.makedirs(state_dir)

        results = querier.async_wait_for_result(
            querier.seed()
        )
        assert len(results['results']) == 1

        get_write_manifest(querier, os.path.join(state_dir, 'manifest.json'))

        project.seeds['my_seed.csv'] = 'a,b\n1,hello\n2,goodbye'
        project.write_seeds(project_root, remove=True)

        querier.sighup()
        assert querier.wait_for_status('ready') is True

        results = querier.async_wait_for_result(
            querier.seed(state='./state', select=['state:modified'])
        )
        assert len(results['results']) == 1

        get_write_manifest(querier, os.path.join(state_dir, 'manifest.json'))

        results = querier.async_wait_for_result(
            querier.seed(state='./state', select=['state:modified']),
        )
        assert len(results['results']) == 0



import pytest

from .util import (
    get_querier,
    ProjectDefinition,
)

macros_data = '''
{% macro foo() %}
    {{ return(1) }}
{% endmacro %}
{% macro bar(value) %}
    {{ return(value + 1) }}
{% endmacro %}
{% macro quux(value) %}
    {{ return(asdf) }}
{% endmacro %}
'''


@pytest.mark.supported('postgres')
def test_run_operation(
    project_root, profiles_root, dbt_profile, unique_schema
):
    project = ProjectDefinition(
        models={'my_model.sql': 'select 1 as id'},
        macros={
            'my_macros.sql': macros_data,
        }
    )
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )

    with querier_ctx as querier:
        poll_result = querier.async_wait_for_result(
            querier.run_operation(macro='foo', args={})
        )

        assert 'success' in poll_result
        assert poll_result['success'] is True

        poll_result = querier.async_wait_for_result(
            querier.run_operation(macro='bar', args={'value': 10})
        )

        assert 'success' in poll_result
        assert poll_result['success'] is True

        poll_result = querier.async_wait_for_result(
            querier.run_operation(macro='baz', args={}),
            state='failed',
        )
        assert 'state' in poll_result
        assert poll_result['state'] == 'failed'

        poll_result = querier.async_wait_for_result(
            querier.run_operation(macro='quux', args={})
        )
        assert 'success' in poll_result
        assert poll_result['success'] is True


@pytest.mark.supported('postgres')
def test_run_operation_cli(
    project_root, profiles_root, dbt_profile, unique_schema
):
    project = ProjectDefinition(
        models={'my_model.sql': 'select 1 as id'},
        macros={
            'my_macros.sql': macros_data,
        }
    )
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )

    with querier_ctx as querier:
        poll_result = querier.async_wait_for_result(
            querier.cli_args(cli='run-operation foo')
        )

        assert 'success' in poll_result
        assert poll_result['success'] is True

        bar_cmd = '''run-operation bar --args="{'value': 10}"'''
        poll_result = querier.async_wait_for_result(
            querier.cli_args(cli=bar_cmd)
        )

        assert 'success' in poll_result
        assert poll_result['success'] is True

        poll_result = querier.async_wait_for_result(
            querier.cli_args(cli='run-operation baz'),
            state='failed',
        )
        assert 'state' in poll_result
        assert poll_result['state'] == 'failed'

        poll_result = querier.async_wait_for_result(
            querier.cli_args(cli='run-operation quux')
        )
        assert 'success' in poll_result
        assert poll_result['success'] is True




import os
import pytest
import yaml
from .util import (
    assert_has_threads,
    get_querier,
    get_write_manifest,
    ProjectDefinition,
)


@pytest.mark.supported('postgres')
def test_rpc_test_threads(
    project_root, profiles_root, dbt_profile, unique_schema
):
    schema_yaml = {
        'version': 2,
        'models': [{
            'name': 'my_model',
            'columns': [
                {
                    'name': 'id',
                    'tests': ['not_null', 'unique'],
                },
            ],
        }],
    }
    project = ProjectDefinition(
        models={
            'my_model.sql': 'select 1 as id',
            'schema.yml': yaml.safe_dump(schema_yaml)}
    )
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )
    with querier_ctx as querier:
        # first run dbt to get the model built
        querier.async_wait_for_result(querier.run())

        results = querier.async_wait_for_result(querier.test(threads=5))
        assert_has_threads(results, 5)

        results = querier.async_wait_for_result(
            querier.cli_args('test --threads=7')
        )
        assert_has_threads(results, 7)


@pytest.mark.supported('postgres')
def test_rpc_test_state(
    project_root, profiles_root, dbt_profile, unique_schema
):
    schema_yaml = {
        'version': 2,
        'models': [{
            'name': 'my_model',
            'columns': [
                {
                    'name': 'id',
                    'tests': ['not_null', 'unique'],
                },
            ],
        }],
    }
    project = ProjectDefinition(
        models={
            'my_model.sql': 'select 1 as id',
            'schema.yml': yaml.safe_dump(schema_yaml)
        }
    )
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )
    with querier_ctx as querier:
        state_dir = os.path.join(project_root, 'state')
        os.makedirs(state_dir)

        results = querier.async_wait_for_result(querier.run())
        results = querier.async_wait_for_result(querier.test())
        assert len(results['results']) == 2

        get_write_manifest(querier, os.path.join(state_dir, 'manifest.json'))

        project.models['my_model.sql'] = 'select 2 as id'
        project.write_models(project_root, remove=True)
        querier.sighup()
        assert querier.wait_for_status('ready') is True

        results = querier.async_wait_for_result(
            querier.test(state='./state', models=['state:modified'])
        )
        assert len(results['results']) == 2

        get_write_manifest(querier, os.path.join(state_dir, 'manifest.json'))

        results = querier.async_wait_for_result(
            querier.test(state='./state', models=['state:modified']),
        )
        assert len(results['results']) == 0

        # a better test of defer would require multiple targets
        results = querier.async_wait_for_result(
            querier.run(state='./state', models=['state:modified'], defer=True)
        )
        assert len(results['results']) == 0



import json
from typing import Callable, Dict, Any

from dbt.dataclass_schema import dbtClassMixin
from jsonrpc.exceptions import (
    JSONRPCParseError,
    JSONRPCInvalidRequestException,
    JSONRPCInvalidRequest,
)
from jsonrpc import JSONRPCResponseManager
from jsonrpc.jsonrpc import JSONRPCRequest
from jsonrpc.jsonrpc2 import JSONRPC20Request, JSONRPC20Response
from werkzeug import Request as HTTPRequest

import dbt.exceptions
import dbt.tracking
from dbt.logger import GLOBAL_LOGGER as logger
from dbt_rpc.rpc.logger import RequestContext
from dbt_rpc.rpc.task_handler import RequestTaskHandler
from dbt_rpc.rpc.method import RemoteMethod
from dbt_rpc.rpc.task_manager import TaskManager


def track_rpc_request(task):
    dbt.tracking.track_rpc_request({
        "task": task
    })


SYNCHRONOUS_REQUESTS = False


class RequestDispatcher(Dict[str, Callable[..., Dict[str, Any]]]):
    """A special dispatcher that knows about requests."""
    def __init__(
        self,
        http_request: HTTPRequest,
        json_rpc_request: JSONRPC20Request,
        manager: TaskManager,
    ):
        self.http_request = http_request
        self.json_rpc_request = json_rpc_request
        self.manager = manager

    def __getitem__(self, key) -> Callable[..., Dict[str, Any]]:
        handler = self.manager.get_handler(
            key,
            self.http_request,
            self.json_rpc_request,
        )
        if handler is None:
            raise KeyError(key)
        if callable(handler):
            # either an error or a builtin
            return handler
        elif isinstance(handler, RemoteMethod):
            # the handler must be a task. Wrap it in a task handler so it can
            # go async
            return RequestTaskHandler(
                self.manager, handler, self.http_request, self.json_rpc_request
            )
        else:
            raise dbt.exceptions.InternalException(
                f'Got an invalid handler from get_handler. Expected None, '
                f'callable, or RemoteMethod, got {handler}'
            )


class ResponseManager(JSONRPCResponseManager):
    """Override the default response manager to handle request metadata and
    track in-flight tasks via the task manager.
    """
    @classmethod
    def handle_valid_request(
        cls,
        http_request: HTTPRequest,
        request: JSONRPC20Request,
        task_manager: TaskManager,
    ) -> JSONRPC20Response:
        with RequestContext(request):
            logger.info('handling {} request'.format(request.method))
            track_rpc_request(request.method)

            dispatcher = RequestDispatcher(
                http_request, request, task_manager
            )

            return cls.handle_request(request, dispatcher)

    @classmethod
    def _get_responses(cls, requests, dispatcher, context=None):
        for output in super()._get_responses(requests, dispatcher, context):
            # if it's a result, check if it's a dbtClassMixin and if so call
            # to_dict
            if hasattr(output, 'result'):
                if isinstance(output.result, dbtClassMixin):
                    # Note: errors in to_dict do not show up anywhere in
                    # the output and all you get is a generic 500 error
                    output.result = \
                        output.result.to_dict(omit_none=False)
            yield output

    @classmethod
    def handle(
        cls,
        http_request: HTTPRequest,
        task_manager: TaskManager,
    ) -> JSONRPC20Response:
        request_str: str
        if isinstance(http_request.data, bytes):
            request_str = http_request.data.decode("utf-8")
        else:
            request_str = http_request.data

        try:
            data = json.loads(request_str)
        except (TypeError, ValueError):
            return JSONRPC20Response(error=dict(
                code=JSONRPCParseError.CODE,
                message=JSONRPCParseError.MESSAGE,
            ))

        if data.get('jsonrpc', None) != '2.0':
            return JSONRPC20Response(error=dict(
                code=JSONRPCInvalidRequest.CODE,
                message=JSONRPCInvalidRequest.MESSAGE,
            ))

        try:
            request = JSONRPCRequest.from_data(data)
        except (ValueError, JSONRPCInvalidRequestException):
            return JSONRPC20Response(error=dict(
                code=JSONRPCInvalidRequest.CODE,
                message=JSONRPCInvalidRequest.MESSAGE,
            ))

        if not isinstance(request, JSONRPC20Request):
            return JSONRPC20Response(error=dict(
                code=JSONRPCInvalidRequest.CODE,
                message=JSONRPCInvalidRequest.MESSAGE,
            ))

        result = cls.handle_valid_request(
            http_request, request, task_manager
        )

        return result



import pytest

from .util import (
    get_querier,
    ProjectDefinition,
)


@pytest.mark.parametrize(
    "packages, bad_packages",
    # from dbt hub
    [(
        [{
            'package': 'dbt-labs/dbt_utils',
            'version': '0.5.0',
        }],
        # wrong package name
        [{
            'package': 'dbt-labs/dbt_util',
            'version': '0.5.0',
        }],
    ),
    # from git release/tag/branch
    (
        [{
            'git': 'https://github.com/dbt-labs/dbt-utils.git',
            'revision': '0.5.0',
        }],
        # if you use a bad URL, git thinks it's a private repo and prompts for auth
        [{
            'git': 'https://github.com/dbt-labs/dbt-utils.git',
            'revision': 'not-a-real-revision',
        }],
    ),
    # from git commit
    (
        [{
            'git': 'https://github.com/dbt-labs/dbt-utils.git',
            'revision': 'b736cf6acdbf80d2de69b511a51c8d7fe214ee79',
        }],
        # don't use short commits
        [{
            'git': 'https://github.com/dbt-labs/dbt-utils.git',
            'revision': 'b736cf6',
        }],
    ),
    # from git release and subdirectory
    (
        [{
            'git': 'https://github.com/dbt-labs/dbt-labs-experimental-features.git',
            'revision': '0.0.1',
            'subdirectory': 'materialized-views',
        }],
        [{
            'git': 'https://github.com/dbt-labs/dbt-labs-experimental-features.git',
            'revision': '0.0.1',
            'subdirectory': 'path/to/nonexistent/dir',
        }],
    ),
    # from git commit and subdirectory
    (
        [{
            'git': 'https://github.com/dbt-labs/dbt-utils.git',
            'revision': 'f4f84e9110db26aba22f756abbae9f1f8dbb15da',
            'subdirectory': 'dbt_projects/dbt_utils',
        }],
        [{
            'git': 'https://github.com/dbt-labs/dbt-utils.git',
            'revision': 'f4f84e9110db26aba22f756abbae9f1f8dbb15da',
            'subdirectory': 'path/to/nonexistent/dir',
        }],
    )],
    ids=[
        "from dbt hub",
        "from git release/tag/branch",
        "from git commit",
        "from git release and subdirectory",
        "from git commit and subdirectory",
    ],
)
@pytest.mark.supported('postgres')
def test_rpc_deps_packages(project_root, profiles_root, dbt_profile, unique_schema, packages, bad_packages):
    project = ProjectDefinition(
        models={
            'my_model.sql': 'select 1 as id',
        },
        packages={'packages': packages},
    )
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
        criteria='error',
    )
    with querier_ctx as querier:
        # we should be able to run sql queries at startup
        querier.is_error(querier.run_sql('select 1 as id'))

        # the status should be an error as deps wil not be defined
        querier.is_result(querier.status())

        # deps should pass
        querier.async_wait_for_result(querier.deps())

        # queries should work after deps
        tok1 = querier.is_async_result(querier.run())
        tok2 = querier.is_async_result(querier.run_sql('select 1 as id'))

        querier.is_result(querier.async_wait(tok2))
        querier.is_result(querier.async_wait(tok1))

        # now break the project
        project.packages['packages'] = bad_packages
        project.write_packages(project_root, remove=True)

        # queries should still work because we haven't reloaded
        tok1 = querier.is_async_result(querier.run())
        tok2 = querier.is_async_result(querier.run_sql('select 1 as id'))

        querier.is_result(querier.async_wait(tok2))
        querier.is_result(querier.async_wait(tok1))

        # now run deps again, it should be sad
        querier.async_wait_for_error(querier.deps())
        # it should also not be running.
        result = querier.is_result(querier.ps(active=True, completed=False))
        assert result['rows'] == []

        # fix packages again
        project.packages['packages'] = packages
        project.write_packages(project_root, remove=True)
        # keep queries broken, we haven't run deps yet
        querier.is_error(querier.run())

        # deps should pass now
        querier.async_wait_for_result(querier.deps())
        querier.is_result(querier.status())

        tok1 = querier.is_async_result(querier.run())
        tok2 = querier.is_async_result(querier.run_sql('select 1 as id'))

        querier.is_result(querier.async_wait(tok2))
        querier.is_result(querier.async_wait(tok1))


@pytest.mark.supported('postgres')
def test_rpc_deps_after_list(project_root, profiles_root, dbt_profile, unique_schema):
    packages = [{
        'package': 'dbt-labs/dbt_utils',
        'version': '0.5.0',
    }]
    project = ProjectDefinition(
        models={
            'my_model.sql': 'select 1 as id',
        },
        packages={'packages': packages},
    )
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
        criteria='error',
    )
    with querier_ctx as querier:
        # we should be able to run sql queries at startup
        querier.is_error(querier.run_sql('select 1 as id'))

        # the status should be an error as deps wil not be defined
        querier.is_result(querier.status())

        # deps should pass
        querier.async_wait_for_result(querier.deps())

        # queries should work after deps
        tok1 = querier.is_async_result(querier.run())
        tok2 = querier.is_async_result(querier.run_sql('select 1 as id'))

        querier.is_result(querier.async_wait(tok2))
        querier.is_result(querier.async_wait(tok1))

        # list should pass
        querier.list()

        # deps should pass
        querier.async_wait_for_result(querier.deps())



import pytest
import time
from .util import (
    get_querier,
    ProjectDefinition,
)


@pytest.mark.supported('postgres')
def test_rpc_basics(
    project_root, profiles_root, dbt_profile, unique_schema
):
    project = ProjectDefinition(
        models={'my_model.sql': 'select 1 as id'}
    )
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )

    with querier_ctx as querier:
        querier.async_wait_for_result(querier.run_sql('select 1 as id'))

        querier.async_wait_for_result(querier.run())

        querier.async_wait_for_result(
            querier.run_sql('select * from {{ ref("my_model") }}')
        )

        querier.async_wait_for_error(
            querier.run_sql('select * from {{ reff("my_model") }}')
        )


bad_schema_yml = '''
version: 2
sources:
  - name: test_source
    loader: custom
    schema: "{{ var('test_run_schema') }}"
    tables:
      - name: test_table
        identifier: source
        tests:
          - relationships:
            # this is invalid
              - column_name: favorite_color
              - to: ref('descendant_model')
              - field: favorite_color
'''

fixed_schema_yml = '''
version: 2
sources:
  - name: test_source
    loader: custom
    schema: "{{ var('test_run_schema') }}"
    tables:
      - name: test_table
        identifier: source
'''


@pytest.mark.supported('postgres')
def test_rpc_status_error(project_root, profiles_root, dbt_profile, unique_schema):
    project = ProjectDefinition(
        models={
            'descendant_model.sql': 'select * from {{ source("test_source", "test_table") }}',
            'schema.yml': bad_schema_yml,
        }
    )
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
        criteria='error',
    )
    with querier_ctx as querier:

        # the status should be an error result
        result = querier.is_result(querier.status())
        assert 'error' in result
        assert 'message' in result['error']
        assert 'Invalid test config' in result['error']['message']
        assert 'state' in result
        assert result['state'] == 'error'
        assert 'logs' in result
        logs = result['logs']
        assert len(logs) > 0
        for key in ('message', 'timestamp', 'levelname', 'level'):
            assert key in logs[0]
        assert 'pid' in result
        assert querier.server.pid == result['pid']

        error = querier.is_error(querier.compile_sql('select 1 as id'))
        assert 'code' in error
        assert error['code'] == 10011
        assert 'message' in error
        assert error['message'] == 'RPC server failed to compile project, call the "status" method for compile status'
        assert 'data' in error
        assert 'message' in error['data']
        assert 'Invalid test config' in error['data']['message']

        # deps should fail because it still can't parse the manifest
        querier.async_wait_for_error(querier.deps())

        # and not resolve the issue
        result = querier.is_result(querier.status())
        assert 'error' in result
        assert 'message' in result['error']
        assert 'Invalid test config' in result['error']['message']

        error = querier.is_error(querier.compile_sql('select 1 as id'))
        assert 'code' in error
        assert error['code'] == 10011

        project.models['schema.yml'] = fixed_schema_yml
        project.write_models(project_root, remove=True)

        # deps should work
        querier.async_wait_for_result(querier.deps())

        result = querier.is_result(querier.status())
        assert result.get('error') is None
        assert 'state' in result
        assert result['state'] == 'ready'

        querier.is_result(querier.compile_sql('select 1 as id'))


@pytest.mark.supported('postgres')
def test_gc_change_interval(project_root, profiles_root, dbt_profile, unique_schema):
    project = ProjectDefinition(
        models={'my_model.sql': 'select 1 as id'}
    )
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )

    with querier_ctx as querier:

        for _ in range(10):
            querier.async_wait_for_result(querier.run())

        result = querier.is_result(querier.ps(True, True))
        assert len(result['rows']) == 10

        result = querier.is_result(querier.gc(settings=dict(maxsize=1000, reapsize=5, auto_reap_age=0.1)))

        for k in ('deleted', 'missing', 'running'):
            assert k in result
            assert len(result[k]) == 0

        time.sleep(0.5)

        result = querier.is_result(querier.ps(True, True))
        assert len(result['rows']) == 0

        result = querier.is_result(querier.gc(settings=dict(maxsize=2, reapsize=5, auto_reap_age=100000)))
        for k in ('deleted', 'missing', 'running'):
            assert k in result
            assert len(result[k]) == 0

        time.sleep(0.5)

        for _ in range(10):
            querier.async_wait_for_result(querier.run())

        time.sleep(0.5)
        result = querier.is_result(querier.ps(True, True))
        assert len(result['rows']) <= 2


@pytest.mark.supported('postgres')
def test_ps_poll_output_match(project_root, profiles_root, dbt_profile, unique_schema):
    project = ProjectDefinition(
        models={'my_model.sql': 'select 1 as id'}
    )
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )

    with querier_ctx as querier:

        poll_result = querier.async_wait_for_result(querier.run())

        result = querier.is_result(querier.ps(active=True, completed=True))
        assert 'rows' in result
        rows = result['rows']
        assert len(rows) == 1
        ps_result = rows[0]

        for key in ('start', 'end', 'elapsed', 'state'):
            assert ps_result[key] == poll_result[key]


sleeper_sql = '''
{{ log('test output', info=True) }}
{{ run_query('select * from pg_sleep(20)') }}
select 1 as id
'''

logger_sql = '''
{{ log('test output', info=True) }}
select 1 as id
'''


def find_log_ordering(logs, *messages) -> bool:
    log_iter = iter(logs)
    found = 0

    while found < len(messages):
        try:
            log = next(log_iter)
        except StopIteration:
            return False
        if messages[found] in log['message']:
            found += 1
    return True


def poll_logs(querier, token):
    has_log = querier.is_result(querier.poll(token))
    assert 'logs' in has_log
    return has_log['logs']


def wait_for_log_ordering(querier, token, attempts, *messages) -> int:
    for _ in range(attempts):
        time.sleep(1)
        logs = poll_logs(querier, token)
        if find_log_ordering(logs, *messages):
            return len(logs)

    msg = 'Never got expected messages {} in {}'.format(
        messages,
        [log['message'] for log in logs],
    )
    assert False, msg


@pytest.mark.supported('postgres')
def test_get_status(
    project_root, profiles_root, dbt_profile, unique_schema
):
    project = ProjectDefinition(
        models={'my_model.sql': 'select 1 as id'},
    )
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )

    with querier_ctx as querier:
        # make sure that logs_start/logs are honored during a task
        token = querier.is_async_result(querier.run_sql(sleeper_sql))

        no_log = querier.is_result(querier.poll(token, logs=False))
        assert 'logs' in no_log
        assert len(no_log['logs']) == 0

        num_logs = wait_for_log_ordering(querier, token, 10)

        trunc_log = querier.is_result(querier.poll(token, logs_start=num_logs))
        assert 'logs' in trunc_log
        assert len(trunc_log['logs']) == 0

        querier.kill(token)

        # make sure that logs_start/logs are honored after a task has finished
        token = querier.is_async_result(querier.run_sql(logger_sql))
        result = querier.is_result(querier.async_wait(token))
        assert 'logs' in result
        num_logs = len(result['logs'])
        assert num_logs > 0

        result = querier.is_result(querier.poll(token, logs_start=num_logs))
        assert 'logs' in result
        assert len(result['logs']) == 0

        result = querier.is_result(querier.poll(token, logs=False))
        assert 'logs' in result
        assert len(result['logs']) == 0


@pytest.mark.supported('postgres')
def test_missing_tag_sighup(
    project_root, profiles_root, dbt_profile, unique_schema
):
    project = ProjectDefinition(
        models={
            'my_docs.md': '{% docs asdf %}have a close tag{% enddocs %}',
        },
    )
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )
    with querier_ctx as querier:
        # everything is fine
        assert querier.wait_for_status('ready') is True

        # write a junk docs file
        project.models['my_docs.md'] = '{% docs asdf %}do not have a close tag'
        project.write_models(project_root, remove=True)

        querier.sighup()

        assert querier.wait_for_status('error') is True
        result = querier.is_result(querier.status())
        assert 'error' in result
        assert 'message' in result['error']
        assert 'without finding a close tag for docs' in result['error']['message']

        project.models['my_docs.md'] = '{% docs asdf %}have a close tag again{% enddocs %}'
        project.write_models(project_root, remove=True)

        querier.sighup()

        assert querier.wait_for_status('ready') is True


@pytest.mark.supported('postgres')
def test_get_manifest(
    project_root, profiles_root, dbt_profile, unique_schema
):
    project = ProjectDefinition(
        models={
            'my_model.sql': 'select 1 as id',
        },
    )
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )

    with querier_ctx as querier:
        results = querier.async_wait_for_result(querier.cli_args('run'))
        assert len(results['results']) == 1
        assert results['results'][0]['node']['compiled_sql'] == 'select 1 as id'
        result = querier.async_wait_for_result(querier.get_manifest())
        assert 'manifest' in result
        # manifest is not used for IDE, we are sending back the raw manifest.
        manifest = result['manifest']
        assert manifest['nodes']['model.test.my_model']['raw_code'] == 'select 1 as id'
        assert 'manifest' in result
        manifest = result['manifest']
        assert manifest['nodes']['model.test.my_model']['compiled_code'] == 'select 1 as id'


@pytest.mark.supported('postgres')
def test_variable_injection(
    project_root, profiles_root, dbt_profile, unique_schema
):
    project = ProjectDefinition(
        models={
            'my_model.sql': 'select {{ var("test_variable") }} as id',
        },
    )
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )

    with querier_ctx as querier:
      results = querier.async_wait_for_result(querier.cli_args('run --vars \'test_variable: "1234"\''))
      assert len(results['results']) == 1
      assert results['results'][0]['node']['compiled_sql'] == 'select 1234 as id'


config_var_sql = '''
{{ config(unique_key=var('test_variable')) }}
select {{ config.get('unique_key') }} as id
'''


@pytest.mark.supported('postgres')
def test_variable_injection_in_config(
    project_root, profiles_root, dbt_profile, unique_schema
):
    project = ProjectDefinition(
        models={
            'my_model.sql': config_var_sql,
        },
    )
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )

    with querier_ctx as querier:
      results = querier.async_wait_for_result(querier.cli_args('run --vars \'test_variable: "9876"\''))
      assert len(results['results']) == 1
      assert results['results'][0]['node']['compiled_sql'].strip() == 'select 9876 as id'



from typing import List, Dict, Any, Optional

from jsonrpc.exceptions import JSONRPCDispatchException, JSONRPCInvalidParams

import dbt.exceptions


class RPCException(JSONRPCDispatchException):
    def __init__(
        self,
        code: Optional[int] = None,
        message: Optional[str] = None,
        data: Optional[Dict[str, Any]] = None,
        logs: Optional[List[Dict[str, Any]]] = None,
        tags: Optional[Dict[str, Any]] = None
    ) -> None:
        if code is None:
            code = -32000
        if message is None:
            message = 'Server error'
        if data is None:
            data = {}

        super().__init__(code=code, message=message, data=data)
        if logs is not None:
            self.logs = logs
        self.error.data['tags'] = tags

    def __str__(self):
        return (
            'RPCException({0.code}, {0.message}, {0.data}, {1.logs})'
            .format(self.error, self)
        )

    @property
    def logs(self) -> List[Dict[str, Any]]:
        return self.error.data.get('logs')

    @logs.setter
    def logs(self, value):
        if value is None:
            return
        self.error.data['logs'] = value

    @property
    def tags(self):
        return self.error.data.get('tags')

    @tags.setter
    def tags(self, value):
        if value is None:
            return
        self.error.data['tags'] = value

    @classmethod
    def from_error(cls, err):
        return cls(
            code=err.code,
            message=err.message,
            data=err.data,
            logs=err.data.get('logs'),
            tags=err.data.get('tags'),
        )


def invalid_params(data):
    return RPCException(
        code=JSONRPCInvalidParams.CODE,
        message=JSONRPCInvalidParams.MESSAGE,
        data=data
    )


def server_error(err, logs=None, tags=None):
    exc = dbt.exceptions.Exception(str(err))
    return dbt_error(exc, logs, tags)


def timeout_error(timeout_value, logs=None, tags=None):
    exc = dbt.exceptions.RPCTimeoutException(timeout_value)
    return dbt_error(exc, logs, tags)


def dbt_error(exc, logs=None, tags=None):
    exc = RPCException(code=exc.CODE, message=exc.MESSAGE, data=exc.data(),
                       logs=logs, tags=tags)
    return exc



import os
import pytest
from .util import (
    assert_has_threads,
    get_querier,
    get_write_manifest,
    ProjectDefinition,
)

snapshot_data = '''
{% snapshot snapshot_actual %}

    {{
        config(
            target_database=database,
            target_schema=schema,
            unique_key='id',
            strategy='timestamp',
            updated_at='updated_at',
        )
    }}
    select 1 as id, cast('2019-10-31 23:59:40' as timestamp) as updated_at

{% endsnapshot %}
'''

snapshot_data_2 = '''
{% snapshot snapshot_actual %}

    {{
        config(
            target_database=database,
            target_schema=schema,
            unique_key='id',
            strategy='timestamp',
            updated_at='updated_at',
        )
    }}
    select 2 as id, cast('2019-10-31 23:59:40' as timestamp) as updated_at

{% endsnapshot %}
'''


@pytest.mark.supported('postgres')
def test_snapshots(
    project_root, profiles_root, dbt_profile, unique_schema
):
    project = ProjectDefinition(
        snapshots={'my_snapshots.sql': snapshot_data},
    )
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )

    with querier_ctx as querier:
        results = querier.async_wait_for_result(querier.snapshot())
        assert len(results['results']) == 1

        results = querier.async_wait_for_result(querier.snapshot(
            exclude=['snapshot_actual'])
        )

        results = querier.async_wait_for_result(
            querier.snapshot(select=['snapshot_actual'])
        )
        assert len(results['results']) == 1


@pytest.mark.supported('postgres')
def test_snapshots_cli(
    project_root, profiles_root, dbt_profile, unique_schema
):
    project = ProjectDefinition(
        snapshots={'my_snapshots.sql': snapshot_data},
    )
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )

    with querier_ctx as querier:
        results = querier.async_wait_for_result(
            querier.cli_args(cli='snapshot')
        )
        assert len(results['results']) == 1

        results = querier.async_wait_for_result(
            querier.cli_args(cli='snapshot --exclude=snapshot_actual')
        )
        assert len(results['results']) == 0

        results = querier.async_wait_for_result(
            querier.cli_args(cli='snapshot --select=snapshot_actual')
        )
        assert len(results['results']) == 1


@pytest.mark.supported('postgres')
def test_rpc_snapshot_threads(
    project_root, profiles_root, dbt_profile, unique_schema
):
    project = ProjectDefinition(
        snapshots={'my_snapshots.sql': snapshot_data},
    )
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )

    with querier_ctx as querier:
        results = querier.async_wait_for_result(querier.snapshot(threads=5))
        assert_has_threads(results, 5)

        results = querier.async_wait_for_result(
            querier.cli_args('snapshot --threads=7')
        )
        assert_has_threads(results, 7)


@pytest.mark.supported('postgres')
def test_rpc_snapshot_state(
    project_root, profiles_root, dbt_profile, unique_schema
):
    project = ProjectDefinition(
        snapshots={'my_snapshots.sql': snapshot_data},
    )
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )
    with querier_ctx as querier:
        state_dir = os.path.join(project_root, 'state')
        os.makedirs(state_dir)

        results = querier.async_wait_for_result(
            querier.snapshot()
        )
        assert len(results['results']) == 1

        get_write_manifest(querier, os.path.join(state_dir, 'manifest.json'))

        project.snapshots['my_snapshots.sql'] = snapshot_data_2
        project.write_snapshots(project_root, remove=True)

        querier.sighup()
        assert querier.wait_for_status('ready') is True

        results = querier.async_wait_for_result(
            querier.snapshot(state='./state', select=['state:modified'])
        )
        assert len(results['results']) == 1

        get_write_manifest(querier, os.path.join(state_dir, 'manifest.json'))

        results = querier.async_wait_for_result(
            querier.snapshot(state='./state', select=['state:modified']),
        )
        assert len(results['results']) == 0



import os
import pytest
import yaml
from .util import (
    assert_has_threads,
    get_querier,
    get_write_manifest,
    ProjectDefinition,
)

snapshot_data = '''
{% snapshot my_snapshot %}

    {{
        config(
            target_database=database,
            target_schema=schema,
            unique_key='id',
            strategy='timestamp',
            updated_at='updated_at',
        )
    }}
    select id, cast('2019-10-31 23:59:40' as timestamp) as updated_at
    from {{ ref('my_model') }}

{% endsnapshot %}
'''


@pytest.mark.supported('postgres')
def test_rpc_build_threads(
    project_root, profiles_root, dbt_profile, unique_schema
):
    schema_yaml = {
        'version': 2,
        'models': [{
            'name': 'my_model',
            'columns': [
                {
                    'name': 'id',
                    'tests': ['not_null', 'unique'],
                },
            ],
        }],
    }
    project = ProjectDefinition(
        project_data={'seeds': {'+quote_columns': False}},
        models={
            'my_model.sql': 'select * from {{ ref("data") }}',
            'schema.yml': yaml.safe_dump(schema_yaml)
        },
        seeds={'data.csv': 'id,message\n1,hello\n2,goodbye'},
        snapshots={'my_snapshots.sql': snapshot_data},
    )
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )
    with querier_ctx as querier:
        # first run dbt to get the model built
        querier.async_wait_for_result(querier.build())

        results = querier.async_wait_for_result(querier.test(threads=5))
        assert_has_threads(results, 5)

        results = querier.async_wait_for_result(
            querier.cli_args('build --threads=7')
        )
        assert_has_threads(results, 7)


@pytest.mark.supported('postgres')
def test_rpc_build_state(
    project_root, profiles_root, dbt_profile, unique_schema
):
    schema_yaml = {
        'version': 2,
        'models': [{
            'name': 'my_model',
            'columns': [
                {
                    'name': 'id',
                    'tests': ['not_null', 'unique'],
                },
            ],
        }],
    }
    project = ProjectDefinition(
        project_data={'seeds': {'+quote_columns': False}},
        models={
            'my_model.sql': 'select * from {{ ref("data") }}',
            'schema.yml': yaml.safe_dump(schema_yaml)
        },
        seeds={'data.csv': 'id,message\n1,hello\n2,goodbye'},
        snapshots={'my_snapshots.sql': snapshot_data},
    )
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )
    with querier_ctx as querier:
        state_dir = os.path.join(project_root, 'state')
        os.makedirs(state_dir)

        results = querier.async_wait_for_result(querier.build())
        assert len(results['results']) == 5

        get_write_manifest(querier, os.path.join(state_dir, 'manifest.json'))

        project.models['my_model.sql'] =\
            'select * from {{ ref("data" )}} where id = 2'
        project.write_models(project_root, remove=True)
        querier.sighup()
        assert querier.wait_for_status('ready') is True

        results = querier.async_wait_for_result(
            querier.build(state='./state', select=['state:modified'])
        )
        assert len(results['results']) == 3

        get_write_manifest(querier, os.path.join(state_dir, 'manifest.json'))

        results = querier.async_wait_for_result(
            querier.build(state='./state', select=['state:modified']),
        )
        assert len(results['results']) == 0

        # a better test of defer would require multiple targets
        results = querier.async_wait_for_result(
            querier.build(
                state='./state',
                select=['state:modified'],
                defer=True
            )
        )
        assert len(results['results']) == 0


@pytest.mark.supported('postgres')
def test_rpc_build_selectors(
    project_root, profiles_root, dbt_profile, unique_schema
):
    schema_yaml = {
        'version': 2,
        'models': [{
            'name': 'my_model',
            'columns': [
                {
                    'name': 'id',
                    'tests': ['not_null', 'unique'],
                },
            ],
        }],
    }
    project = ProjectDefinition(
        name='test',
        project_data={
            'seeds': {'+quote_columns': False},
            'models': {'test': {'my_model': {'+tags': 'example_tag'}}}
        },
        models={
            'my_model.sql': 'select * from {{ ref("data") }}',
            'schema.yml': yaml.safe_dump(schema_yaml)
        },
        seeds={'data.csv': 'id,message\n1,hello\n2,goodbye'},
        snapshots={'my_snapshots.sql': snapshot_data},
    )
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )
    with querier_ctx as querier:
        # test simple resource_types param
        results = querier.async_wait_for_result(
            querier.build(resource_types=['seed'])
        )
        assert len(results['results']) == 1
        assert results['results'][0]['node']['resource_type'] == 'seed'

        # test simple select param (should select tagged model and its tests)
        results = querier.async_wait_for_result(
            querier.build(select=['tag:example_tag'])
        )
        assert len(results['results']) == 3
        assert sorted(
            [result['node']['resource_type'] for result in results['results']]
        ) == ['model', 'test', 'test']



from concurrent.futures import ThreadPoolExecutor, as_completed
import pytest

from .util import (
    get_querier,
    ProjectDefinition,
)


def _compile_poll_for_result(querier, id: int):
    sql = f'select {id} as id'
    resp = querier.compile_sql(
        request_id=id, sql=sql, name=f'query_{id}'
    )
    compile_sql_result = querier.async_wait_for_result(resp)
    assert compile_sql_result['results'][0]['compiled_sql'] == sql


@pytest.mark.supported('postgres')
def test_rpc_compile_sql_concurrency(
    project_root, profiles_root, dbt_profile, unique_schema
):
    project = ProjectDefinition(
        models={'my_model.sql': 'select 1 as id'}
    )
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )

    with querier_ctx as querier:
        values = {}
        with ThreadPoolExecutor(max_workers=10) as tpe:
            for id in range(20):
                fut = tpe.submit(_compile_poll_for_result, querier, id)
                values[fut] = id
            for fut in as_completed(values):
                fut.result()



import os
import pytest
from .util import (
    assert_has_threads,
    get_querier,
    get_write_manifest,
    ProjectDefinition,
)


@pytest.mark.supported('postgres')
def test_rpc_run_threads(
    project_root, profiles_root, dbt_profile, unique_schema
):
    project = ProjectDefinition(
        models={'my_model.sql': 'select 1 as id'}
    )
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )
    with querier_ctx as querier:
        results = querier.async_wait_for_result(querier.run(threads=5))
        assert_has_threads(results, 5)

        results = querier.async_wait_for_result(
            querier.cli_args('run --threads=7')
        )
        assert_has_threads(results, 7)


@pytest.mark.supported('postgres')
def test_rpc_run_vars(
    project_root, profiles_root, dbt_profile, unique_schema
):
    project = ProjectDefinition(
        models={
            'my_model.sql': 'select {{ var("param") }} as id',
        },
    )
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )

    with querier_ctx as querier:
        results = querier.async_wait_for_result(querier.cli_args('run --vars "{param: 100}"'))
        assert len(results['results']) == 1
        assert results['results'][0]['node']['compiled_sql'] == 'select 100 as id'


@pytest.mark.supported('postgres')
def test_rpc_run_vars_compiled(
    project_root, profiles_root, dbt_profile, unique_schema
):
    project = ProjectDefinition(
        models={
            'my_model.sql': '{{ config(materialized=var("materialized_var", "view")) }} select 1 as id',
        },
    )

    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )
    with querier_ctx as querier:
        results = querier.async_wait_for_result(querier.cli_args('--no-partial-parse run --vars "{materialized_var: table}"'))
        assert len(results['results']) == 1
        assert results['results'][0]['node']['config']['materialized'] == 'table'
        # make sure that `--vars` doesn't update global state - if it does,
        # this run() will result in a view!
        results = querier.async_wait_for_result(querier.cli_args('--no-partial-parse run'))
        assert len(results['results']) == 1
        assert results['results'][0]['node']['config']['materialized'] == 'view'


@pytest.mark.supported('postgres')
def test_rpc_run_state_defer(
    project_root, profiles_root, dbt_profile, unique_schema
):
    project = ProjectDefinition(models={'my_model.sql': 'select 1 as id'})
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )
    with querier_ctx as querier:
        state_dir = os.path.join(project_root, 'state')
        os.makedirs(state_dir)

        results = querier.async_wait_for_result(
            querier.run()
        )
        assert len(results['results']) == 1

        get_write_manifest(querier, os.path.join(state_dir, 'manifest.json'))

        project.models['my_model.sql'] = 'select 2 as id'
        project.write_models(project_root, remove=True)
        querier.sighup()
        assert querier.wait_for_status('ready') is True

        results = querier.async_wait_for_result(
            querier.run(state='./state', models=['state:modified'])
        )
        assert len(results['results']) == 1

        get_write_manifest(querier, os.path.join(state_dir, 'manifest.json'))

        results = querier.async_wait_for_result(
            querier.run(state='./state', models=['state:modified']),
        )
        assert len(results['results']) == 0

        project.models['my_model.sql'] = '{% if execute %}{% do exceptions.raise_compiler_error("should not see this") %}{% endif %}select 2 as id'
        project.models['my_second_model.sql'] = 'select * from {{ ref("my_model") }}'
        project.write_models(project_root, remove=True)
        querier.sighup()
        assert querier.wait_for_status('ready') is True

        # if 'defer' is ignored, this will fail
        results = querier.async_wait_for_result(
            querier.run(state='./state', models=['my_second_model'], defer=True)
        )
        assert len(results['results']) == 1



#!/usr/bin/env python
import os
from setuptools import setup, find_namespace_packages


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


package_name = "dbt-rpc"
package_version = "0.4.0"
description = """ A JSON RPC server that provides an interface to programmically interact with dbt projects. """


setup(
    name=package_name,
    version=package_version,
    description=description,
    long_description=description,
    author="dbt Labs",
    author_email="info@dbtlabs.com",
    url="https://github.com/dbt-labs/dbt-rpc",
    packages=find_namespace_packages(include=["dbt_rpc", "dbt_rpc.*"]),
    include_package_data=True,
    test_suite="tests",
    entry_points={
        "console_scripts": [
            "dbt-rpc = dbt_rpc.__main__:main",
        ],
    },
    install_requires=["json-rpc>=1.14,<2", "dbt-core>=1.5.0"],
    zip_safe=False,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    python_requires=">=3.6.3",
)



import pytest
from datetime import datetime, timedelta
from .util import (
    get_querier,
    ProjectDefinition,
)

source_freshness_schema_yml = '''
version: 2
sources:
  - name: test_source
    loaded_at_field: b
    schema: {schema}
    freshness:
      warn_after: {{count: 10, period: hour}}
      error_after: {{count: 1, period: day}}
    tables:
      - name: test_table
        identifier: source
      - name: failure_table
        identifier: other_source
'''

@pytest.mark.supported('postgres')
def test_source_snapshot_freshness(
    project_root, profiles_root, dbt_profile, unique_schema
):
    start_time = datetime.utcnow()
    warn_me = start_time - timedelta(hours=18)
    error_me = start_time - timedelta(days=2)
    # this should trigger a 'warn'
    project = ProjectDefinition(
        project_data={'seeds': {'config': {'quote_columns': False}}},
        seeds={
            'source.csv': 'a,b\n1,{}\n'.format(error_me.strftime('%Y-%m-%d %H:%M:%S')),
            'other_source.csv': 'a,b\n1,{}\n'.format(error_me.strftime('%Y-%m-%d %H:%M:%S'))
        },
        models={
            'sources.yml': source_freshness_schema_yml.format(schema=unique_schema),
        },
    )
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )

    with querier_ctx as querier:
        seeds = querier.async_wait_for_result(querier.seed())
        assert len(seeds['results']) == 2
        # should error
        error_results = querier.async_wait_for_result(querier.snapshot_freshness(), state='failed')
        assert len(error_results['results']) == 2
        for result in error_results['results']:
            assert result['status'] == 'error'
        error_results = querier.async_wait_for_result(querier.cli_args('source snapshot-freshness'), state='failed')
        assert len(error_results['results']) == 2
        for result in error_results['results']:
            assert result['status'] == 'error'

        project.seeds['source.csv'] += '2,{}\n'.format(warn_me.strftime('%Y-%m-%d %H:%M:%S'))
        project.write_seeds(project_root, remove=True)
        querier.async_wait_for_result(querier.seed())
        # should warn
        warn_results = querier.async_wait_for_result(querier.snapshot_freshness(select='source:test_source.test_table'))
        assert len(warn_results['results']) == 1
        assert warn_results['results'][0]['status'] == 'warn'
        warn_results = querier.async_wait_for_result(querier.cli_args('source snapshot-freshness -s source:test_source.test_table'))
        assert len(warn_results['results']) == 1
        assert warn_results['results'][0]['status'] == 'warn'

        project.seeds['source.csv'] += '3,{}\n'.format(start_time.strftime('%Y-%m-%d %H:%M:%S'))
        project.write_seeds(project_root, remove=True)
        querier.async_wait_for_result(querier.seed())
        # should pass!
        pass_results = querier.async_wait_for_result(querier.snapshot_freshness(select=['source:test_source.test_table']))
        assert len(pass_results['results']) == 1
        assert pass_results['results'][0]['status'] == 'pass'
        pass_results = querier.async_wait_for_result(querier.cli_args('source snapshot-freshness --select source:test_source.test_table'))
        assert len(pass_results['results']) == 1
        assert pass_results['results'][0]['status'] == 'pass'

@pytest.mark.supported('postgres')
def test_source_freshness(
    project_root, profiles_root, dbt_profile, unique_schema
):
    start_time = datetime.utcnow()
    warn_me = start_time - timedelta(hours=18)
    error_me = start_time - timedelta(days=2)
    # this should trigger a 'warn'
    project = ProjectDefinition(
        project_data={'seeds': {'config': {'quote_columns': False}}},
        seeds={
            'source.csv': 'a,b\n1,{}\n'.format(error_me.strftime('%Y-%m-%d %H:%M:%S')),
            'other_source.csv': 'a,b\n1,{}\n'.format(error_me.strftime('%Y-%m-%d %H:%M:%S'))
        },
        models={
            'sources.yml': source_freshness_schema_yml.format(schema=unique_schema),
        },
    )
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )

    with querier_ctx as querier:
        seeds = querier.async_wait_for_result(querier.seed())
        assert len(seeds['results']) == 2
        # should error
        error_results = querier.async_wait_for_result(querier.source_freshness(), state='failed')
        assert len(error_results['results']) == 2
        for result in error_results['results']:
            assert result['status'] == 'error'
        error_results = querier.async_wait_for_result(querier.cli_args('source freshness'), state='failed')
        assert len(error_results['results']) == 2
        for result in error_results['results']:
            assert result['status'] == 'error'

        project.seeds['source.csv'] += '2,{}\n'.format(warn_me.strftime('%Y-%m-%d %H:%M:%S'))
        project.write_seeds(project_root, remove=True)
        querier.async_wait_for_result(querier.seed())
        # should warn
        warn_results = querier.async_wait_for_result(querier.source_freshness(select='source:test_source.test_table'))
        assert len(warn_results['results']) == 1
        assert warn_results['results'][0]['status'] == 'warn'
        warn_results = querier.async_wait_for_result(querier.cli_args('source freshness -s source:test_source.test_table'))
        assert len(warn_results['results']) == 1
        assert warn_results['results'][0]['status'] == 'warn'

        project.seeds['source.csv'] += '3,{}\n'.format(start_time.strftime('%Y-%m-%d %H:%M:%S'))
        project.write_seeds(project_root, remove=True)
        querier.async_wait_for_result(querier.seed())
        # should pass!
        pass_results = querier.async_wait_for_result(querier.source_freshness(select=['source:test_source.test_table']))
        assert len(pass_results['results']) == 1
        assert pass_results['results'][0]['status'] == 'pass'
        pass_results = querier.async_wait_for_result(querier.cli_args('source freshness --select source:test_source.test_table'))
        assert len(pass_results['results']) == 1
        assert pass_results['results'][0]['status'] == 'pass'



import json
from datetime import datetime
from pathlib import Path
from typing import List, Optional, Union

from dbt.flags import get_flags
from dbt.contracts.graph.manifest import WritableManifest
from dbt_rpc.contracts.rpc import (
    GetManifestParameters,
    GetManifestResult,
    RPCCompileParameters,
    RPCDocsGenerateParameters,
    RPCRunParameters,
    RPCRunOperationParameters,
    RPCSeedParameters,
    RPCTestParameters,
    RemoteCatalogResults,
    RemoteExecutionResult,
    RemoteListResults,
    RemoteRunOperationResult,
    RPCSnapshotParameters,
    RPCSourceFreshnessParameters,
    RPCListParameters,
    RPCBuildParameters,
)
from dbt.exceptions import DbtRuntimeError
from dbt_rpc.rpc.method import (
    Parameters, RemoteManifestMethod
)

from dbt.task.base import BaseTask
from dbt.task.compile import CompileTask
from dbt.task.freshness import FreshnessTask
from dbt.task.generate import GenerateTask
from dbt.task.run import RunTask
from dbt.task.run_operation import RunOperationTask
from dbt.task.seed import SeedTask
from dbt.task.snapshot import SnapshotTask
from dbt.task.test import TestTask
from dbt.task.list import ListTask
from dbt.task.build import BuildTask

from .base import RPCTask
from .cli import HasCLI


class RPCCommandTask(
    RPCTask[Parameters],
    HasCLI[Parameters, RemoteExecutionResult],
    BaseTask,
):
    @staticmethod
    def _listify(
        value: Optional[Union[str, List[str]]]
    ) -> Optional[List[str]]:
        if value is None:
            return None
        elif isinstance(value, str):
            return [value]
        else:
            return value

    def handle_request(self) -> RemoteExecutionResult:
        return self.run()


def state_path(state: Optional[str]) -> Optional[Path]:
    flags = get_flags()
    if state is not None:
        return Path(state)
    elif flags.STATE is not None:
        return Path(flags.STATE)
    else:
        return None


class RemoteCompileProjectTask(
    RPCCommandTask[RPCCompileParameters], CompileTask
):
    METHOD_NAME = 'compile'

    def set_args(self, params: RPCCompileParameters) -> None:
        if params.models:
            self.args.select = self._listify(params.models)
        else:
            self.args.select = self._listify(params.select)
        self.args.exclude = self._listify(params.exclude)
        self.args.selector_name = params.selector
        if params.threads is not None:
            self.args.threads = params.threads

        self.args.state = state_path(params.state)
        self.set_previous_state()


class RemoteRunProjectTask(RPCCommandTask[RPCRunParameters], RunTask):
    METHOD_NAME = 'run'

    def set_args(self, params: RPCRunParameters) -> None:
        if params.models:
            self.args.select = self._listify(params.models)
        else:
            self.args.select = self._listify(params.select)
        self.args.exclude = self._listify(params.exclude)
        self.args.selector_name = params.selector

        if params.threads is not None:
            self.args.threads = params.threads
        if params.defer is None:
            self.args.defer = get_flags().DEFER
        else:
            self.args.defer = params.defer

        self.args.state = state_path(params.state)
        self.set_previous_state()


class RemoteSeedProjectTask(RPCCommandTask[RPCSeedParameters], SeedTask):
    METHOD_NAME = 'seed'

    def set_args(self, params: RPCSeedParameters) -> None:
        # select has an argparse `dest` value of `models`.
        self.args.select = self._listify(params.select)
        self.args.exclude = self._listify(params.exclude)
        self.args.selector_name = params.selector
        if params.threads is not None:
            self.args.threads = params.threads
        self.args.show = params.show

        self.args.state = state_path(params.state)
        self.set_previous_state()


class RemoteTestProjectTask(RPCCommandTask[RPCTestParameters], TestTask):
    METHOD_NAME = 'test'

    def set_args(self, params: RPCTestParameters) -> None:
        if params.models:
            self.args.select = self._listify(params.models)
        else:
            self.args.select = self._listify(params.select)
        self.args.exclude = self._listify(params.exclude)
        self.args.selector_name = params.selector
        self.args.data = params.data
        self.args.schema = params.schema
        if params.threads is not None:
            self.args.threads = params.threads
        if params.defer is None:
            self.args.defer = get_flags().DEFER
        else:
            self.args.defer = params.defer

        self.args.state = state_path(params.state)
        self.set_previous_state()


class RemoteDocsGenerateProjectTask(
    RPCCommandTask[RPCDocsGenerateParameters],
    GenerateTask,
):
    METHOD_NAME = 'docs.generate'

    def set_args(self, params: RPCDocsGenerateParameters) -> None:
        self.args.select = None
        self.args.exclude = None
        self.args.selector_name = None
        self.args.compile = params.compile

        self.args.state = state_path(params.state)

    def get_catalog_results(
        self, nodes, sources, generated_at, compile_results, errors
    ) -> RemoteCatalogResults:
        return RemoteCatalogResults(
            nodes=nodes,
            sources=sources,
            generated_at=datetime.utcnow(),
            _compile_results=compile_results,
            errors=errors,
            logs=[],
        )


class RemoteRunOperationTask(
    RunOperationTask,
    RemoteManifestMethod[RPCRunOperationParameters, RemoteRunOperationResult],
    HasCLI[RPCRunOperationParameters, RemoteRunOperationResult],
):
    METHOD_NAME = 'run-operation'

    def __init__(self, args, config, manifest):
        super().__init__(args, config)
        RemoteManifestMethod.__init__(
            self, args, config, manifest  # type: ignore
        )

    def load_manifest(self):
        # we started out with a manifest!
        pass

    def set_args(self, params: RPCRunOperationParameters) -> None:
        self.args.macro = params.macro
        self.args.args = params.args

    def _get_kwargs(self):
        if isinstance(self.args.args, dict):
            return self.args.args
        else:
            return RunOperationTask._get_kwargs(self)

    def _runtime_initialize(self):
        return RunOperationTask._runtime_initialize(self)

    def handle_request(self) -> RemoteRunOperationResult:
        base = RunOperationTask.run(self)
        result = RemoteRunOperationResult.from_local_result(base=base, logs=[])
        return result

    def interpret_results(self, results):
        return results.success


class RemoteSnapshotTask(RPCCommandTask[RPCSnapshotParameters], SnapshotTask):
    METHOD_NAME = 'snapshot'

    def set_args(self, params: RPCSnapshotParameters) -> None:
        # select has an argparse `dest` value of `models`.
        self.args.select = self._listify(params.select)
        self.args.exclude = self._listify(params.exclude)
        self.args.selector_name = params.selector
        if params.threads is not None:
            self.args.threads = params.threads

        self.args.state = state_path(params.state)
        self.set_previous_state()


class RemoteSourceFreshnessTask(
    RPCCommandTask[RPCSourceFreshnessParameters],
    FreshnessTask
):
    METHOD_NAME = 'source-freshness'

    def set_args(self, params: RPCSourceFreshnessParameters) -> None:
        self.args.select = self._listify(params.select)
        self.args.exclude = self._listify(params.exclude)
        self.args.selector_name = params.selector
        if params.threads is not None:
            self.args.threads = params.threads
        self.args.output = None


class RemoteSourceSnapshotFreshnessTask(
    RemoteSourceFreshnessTask
):
    """ Deprecated task method name, aliases to `source-freshness` """
    METHOD_NAME = 'snapshot-freshness'


# this is a weird and special method.
class GetManifest(
    RemoteManifestMethod[GetManifestParameters, GetManifestResult]
):
    METHOD_NAME = 'get-manifest'

    def set_args(self, params: GetManifestParameters) -> None:
        self.args.select = None
        self.args.exclude = None
        self.args.selector_name = None

    def handle_request(self) -> GetManifestResult:
        task = RemoteCompileProjectTask(self.args, self.config, self.manifest)
        task.handle_request()

        manifest: Optional[WritableManifest] = None
        if task.manifest is not None:
            manifest = task.manifest.writable_manifest()

        return GetManifestResult(
            logs=[],
            manifest=manifest,
        )

    def interpret_results(self, results):
        return results.manifest is not None


class RemoteListTask(
    RPCCommandTask[RPCListParameters], ListTask
):
    METHOD_NAME = 'list'

    def set_args(self, params: RPCListParameters) -> None:
        self.args.output = params.output
        self.args.output_keys = params.output_keys
        self.args.resource_types = self._listify(params.resource_types)
        self.args.models = self._listify(params.models)
        self.args.exclude = self._listify(params.exclude)
        self.args.selector_name = params.selector
        self.args.select = self._listify(params.select)
        self.args.single_threaded = True

        if self.args.models:
            if self.args.select:
                raise DbtRuntimeError(
                    '"models" and "select" are mutually exclusive arguments'
                )
            if self.args.resource_types:
                raise DbtRuntimeError(
                    '"models" and "resource_type" are mutually exclusive '
                    'arguments'
                )

    @staticmethod
    def output_results(results):
        return RemoteListResults(
            output=[json.loads(x) for x in results],
            logs=None
        )


class RemoteBuildProjectTask(RPCCommandTask[RPCBuildParameters], BuildTask):

    METHOD_NAME = 'build'

    def set_args(self, params: RPCBuildParameters) -> None:
        self.args.resource_types = self._listify(params.resource_types)
        self.args.select = self._listify(params.select)
        self.args.exclude = self._listify(params.exclude)
        self.args.selector_name = params.selector

        if params.threads is not None:
            self.args.threads = params.threads
        if params.defer is None:
            self.args.defer = get_flags().DEFER
        else:
            self.args.defer = params.defer

        self.args.state = state_path(params.state)
        self.set_previous_state()



import pytest

from .util import (
    get_querier,
    ProjectDefinition,
)


@pytest.mark.supported('any')
def test_rpc_run_sql_nohang(
    project_root, profiles_root, dbt_profile, unique_schema
):
    project = ProjectDefinition(
        models={'my_model.sql': 'select 1 as id'}
    )
    querier_ctx = get_querier(
        project_def=project,
        project_dir=project_root,
        profiles_dir=profiles_root,
        schema=unique_schema,
    )
    with querier_ctx as querier:
        querier.async_wait_for_result(querier.run_sql('select 1 as id'))
